@@ -1,4 +1,4 @@
-from fontTools.misc.py23 import *
+from fontTools.misc.py23 import Tag, bytesjoin
 from .DefaultTable import DefaultTable
 import sys
 import array
@@ -134,48 +134,38 @@ class OTTableReader(object):
 		offset = self.offset + offset
 		return self.__class__(self.data, self.localState, offset, self.tableTag)
 
-	def readUShort(self):
+	def readValue(self, typecode, staticSize):
 		pos = self.pos
-		newpos = pos + 2
-		value, = struct.unpack(">H", self.data[pos:newpos])
+		newpos = pos + staticSize
+		value, = struct.unpack(f">{typecode}", self.data[pos:newpos])
 		self.pos = newpos
 		return value
 
-	def readUShortArray(self, count):
+	def readUShort(self):
+		return self.readValue("H", staticSize=2)
+
+	def readArray(self, typecode, staticSize, count):
 		pos = self.pos
-		newpos = pos + count * 2
-		value = array.array("H", self.data[pos:newpos])
+		newpos = pos + count * staticSize
+		value = array.array(typecode, self.data[pos:newpos])
 		if sys.byteorder != "big": value.byteswap()
 		self.pos = newpos
 		return value
 
+	def readUShortArray(self, count):
+		return self.readArray("H", staticSize=2, count=count)
+
 	def readInt8(self):
-		pos = self.pos
-		newpos = pos + 1
-		value, = struct.unpack(">b", self.data[pos:newpos])
-		self.pos = newpos
-		return value
+		return self.readValue("b", staticSize=1)
 
 	def readShort(self):
-		pos = self.pos
-		newpos = pos + 2
-		value, = struct.unpack(">h", self.data[pos:newpos])
-		self.pos = newpos
-		return value
+		return self.readValue("h", staticSize=2)
 
 	def readLong(self):
-		pos = self.pos
-		newpos = pos + 4
-		value, = struct.unpack(">l", self.data[pos:newpos])
-		self.pos = newpos
-		return value
+		return self.readValue("l", staticSize=4)
 
 	def readUInt8(self):
-		pos = self.pos
-		newpos = pos + 1
-		value, = struct.unpack(">B", self.data[pos:newpos])
-		self.pos = newpos
-		return value
+		return self.readValue("B", staticSize=1)
 
 	def readUInt24(self):
 		pos = self.pos
@@ -185,11 +175,7 @@ class OTTableReader(object):
 		return value
 
 	def readULong(self):
-		pos = self.pos
-		newpos = pos + 4
-		value, = struct.unpack(">L", self.data[pos:newpos])
-		self.pos = newpos
-		return value
+		return self.readValue("L", staticSize=4)
 
 	def readTag(self):
 		pos = self.pos
@@ -222,14 +208,24 @@ class OTTableWriter(object):
 
 	"""Helper class to gather and assemble data for OpenType tables."""
 
-	def __init__(self, localState=None, tableTag=None):
+	def __init__(self, localState=None, tableTag=None, offsetSize=2):
 		self.items = []
 		self.pos = None
 		self.localState = localState
 		self.tableTag = tableTag
-		self.longOffset = False
+		self.offsetSize = offsetSize
 		self.parent = None
 
+	# DEPRECATED: 'longOffset' is kept as a property for backward compat with old code.
+	# You should use 'offsetSize' instead (2, 3 or 4 bytes).
+	@property
+	def longOffset(self):
+		return self.offsetSize == 4
+
+	@longOffset.setter
+	def longOffset(self, value):
+		self.offsetSize = 4 if value else 2
+
 	def __setitem__(self, name, value):
 		state = self.localState.copy() if self.localState else dict()
 		state[name] = value
@@ -250,7 +246,7 @@ class OTTableWriter(object):
 			if hasattr(item, "getCountData"):
 				l += item.size
 			elif hasattr(item, "getData"):
-				l += 4 if item.longOffset else 2
+				l += item.offsetSize
 			else:
 				l = l + len(item)
 		return l
@@ -264,9 +260,9 @@ class OTTableWriter(object):
 			item = items[i]
 
 			if hasattr(item, "getData"):
-				if item.longOffset:
+				if item.offsetSize == 4:
 					items[i] = packULong(item.pos - pos)
-				else:
+				elif item.offsetSize == 2:
 					try:
 						items[i] = packUShort(item.pos - pos)
 					except struct.error:
@@ -274,6 +270,10 @@ class OTTableWriter(object):
 						overflowErrorRecord = self.getOverflowErrorRecord(item)
 
 						raise OTLOffsetOverflowError(overflowErrorRecord)
+				elif item.offsetSize == 3:
+					items[i] = packUInt24(item.pos - pos)
+				else:
+					raise ValueError(item.offsetSize)
 
 		return bytesjoin(items)
 
@@ -288,7 +288,7 @@ class OTTableWriter(object):
 	def __eq__(self, other):
 		if type(self) != type(other):
 			return NotImplemented
-		return self.longOffset == other.longOffset and self.items == other.items
+		return self.offsetSize == other.offsetSize and self.items == other.items
 
 	def _doneWriting(self, internedTables):
 		# Convert CountData references to data string items
@@ -409,14 +409,17 @@ class OTTableWriter(object):
 
 	# interface for gathering data, as used by table.compile()
 
-	def getSubWriter(self):
-		subwriter = self.__class__(self.localState, self.tableTag)
+	def getSubWriter(self, offsetSize=2):
+		subwriter = self.__class__(self.localState, self.tableTag, offsetSize=offsetSize)
 		subwriter.parent = self # because some subtables have idential values, we discard
 					# the duplicates under the getAllData method. Hence some
 					# subtable writers can have more than one parent writer.
 					# But we just care about first one right now.
 		return subwriter
 
+	def writeValue(self, typecode, value):
+		self.items.append(struct.pack(f">{typecode}", value))
+
 	def writeUShort(self, value):
 		assert 0 <= value < 0x10000, value
 		self.items.append(struct.pack(">H", value))
@@ -531,6 +534,10 @@ def packULong(value):
 	assert 0 <= value < 0x100000000, value
 	return struct.pack(">L", value)
 
+def packUInt24(value):
+	assert 0 <= value < 0x1000000, value
+	return struct.pack(">L", value)[1:]
+
 
 class BaseTable(object):
 
@@ -647,9 +654,15 @@ class BaseTable(object):
 
 	def compile(self, writer, font):
 		self.ensureDecompiled()
+		# TODO Following hack to be removed by rewriting how FormatSwitching tables
+		# are handled.
+		# https://github.com/fonttools/fonttools/pull/2238#issuecomment-805192631
 		if hasattr(self, 'preWrite'):
+			deleteFormat = not hasattr(self, 'Format')
 			table = self.preWrite(font)
+			deleteFormat = deleteFormat and hasattr(self, 'Format')
 		else:
+			deleteFormat = False
 			table = self.__dict__.copy()
 
 		# some count references may have been initialized in a custom preWrite; we set
@@ -733,6 +746,9 @@ class BaseTable(object):
 				if conv.isPropagated:
 					writer[conv.name] = value
 
+		if deleteFormat:
+			del self.Format
+
 	def readFormat(self, reader):
 		pass
 
@@ -822,6 +838,26 @@ class FormatSwitchingBaseTable(BaseTable):
 		BaseTable.toXML(self, xmlWriter, font, attrs, name)
 
 
+class UInt8FormatSwitchingBaseTable(FormatSwitchingBaseTable):
+	def readFormat(self, reader):
+		self.Format = reader.readUInt8()
+
+	def writeFormat(self, writer):
+		writer.writeUInt8(self.Format)
+
+
+formatSwitchingBaseTables = {
+	"uint16": FormatSwitchingBaseTable,
+	"uint8": UInt8FormatSwitchingBaseTable,
+}
+
+def getFormatSwitchingBaseTableClass(formatType):
+	try:
+		return formatSwitchingBaseTables[formatType]
+	except KeyError:
+		raise TypeError(f"Unsupported format type: {formatType!r}")
+
+
 #
 # Support for ValueRecords
 #
